'''
nose test suite for stencil module

To run some specific tests below use;
nosetests -w tests tests/stencil_tests.py:test_series_loader
'''



from nose.tools import *
from turntable import press

import datetime
import pandas as pd



def setup():
    print "SETUP!"

def teardown():
    print "TEAR DOWN!"

def test_basic():
    print "Test basic"


def test_series_loader():
    print "Testing class SeriesLoader"
    
    series_loader = press.SeriesLoader()
    series_loader.one = 'one'
    print series_loader.series
    assert series_loader.one == 'one'
    assert series_loader.series.one == 'one'


def test_record():
	print "Testing class Record"

	record = press.Record(test = 'test')
	assert record.test == 'test'
	assert record.series.test == 'test'


'''
nose test suite for stencil module

To run some specific tests below use;
nosetests -w tests tests/stencil_tests.py:test_series_loader
'''



from nose.tools import *
import time
from turntable import spin

import datetime
import pandas as pd



def setup():
    print "SETUP!"

def teardown():
    print "TEAR DOWN!"

def test_basic():
    print "Test basic"


def test_mapper_series():
    print "Testing mapper.series"
    
    f = lambda x:x**2.
    seq = range(100)
    assert sum(spin.series(seq, f)) == 328350.0

def f(i):
    o = i*i
    time.sleep(1)
    return o


def test_mapper_process():
    print "Testing mapper.process"


    seq = range(16)
    seq = spin.parallel(seq, f)

    assert sum(seq)==1240


def test_mapper_batch():
    print "Testing mapper.batch"


    seq = range(16)
    seq = spin.batch(seq, f)

    assert sum(seq)==1240




# To use this script, simply import it your setup.py file, and use the
# results of get_git_version() as your package version:
#
# from version import *
#
# setup(
#     version=get_git_version(),
#     .
#     .
#     .
# )
#

__all__ = ("get_git_version",)

from subprocess import Popen, PIPE


def call_git_describe(abbrev=6):
    try:
        p = Popen(['git', 'describe', '--abbrev=%d' % abbrev],
                  stdout=PIPE, stderr=PIPE)
        p.stderr.close()
        line = p.stdout.readlines()[0]
        return line.strip()

    except:
        return None


def read_release_version():
    try:
        f = open("RELEASE-VERSION", "r")

        try:
            version = f.readlines()[0]
            return version.strip()

        finally:
            f.close()

    except:
        return None


def write_release_version(version):
    f = open("RELEASE-VERSION", "w")
    f.write("%s\n" % version)
    f.close()


def get_git_version(abbrev=6):
    # Read in the version that's currently in RELEASE-VERSION.
    release_version = read_release_version()

    version = call_git_describe(abbrev)

    # If that doesn't work, fall back on the value that's in RELEASE-VERSION.
    if version is None:
        version = release_version

    # Adapt to PEP 386 compatible versioning scheme
    version = pep_386_adapt(version)

    # If we still don't have anything, that's an error.
    if version is None:
        raise ValueError("Cannot find the version number!")

    # Remove version prefix 'v' from version number
    if version[0] == 'v':
        version = version[1:]

    # If the current version is different from what's in the
    # RELEASE-VERSION file, update the file to be current.

    if version != release_version:
        write_release_version(version)

    return version


def pep_386_adapt(version):
    if version is not None and '-' in version:
        # Adapt git-describe version to be in line with PEP 386
        parts = version.split('-')
        parts[-2] = 'post' + parts[-2]
        version = '.'.join(parts[:-1])

    return version


if __name__ == "__main__":
    print get_git_version()



'''
Provides a collection of methods used across the package or of general utility.

Example
~~~~~~~
>>> from relpy import utils
>>>
>>> string_date = '2014-05-03'
>>> date = utils.date_extract(string_date)

'''

import os
import re
import sys
import shutil
import numpy as np
import pandas as pd
try:
    import cPickle as pickle
except:
    import pickle
import fnmatch
import errno
import time


def readfile(filename):
    '''
    Checks the extension on the filename and reads the file with the appropriate method.
    *readfile* supports 'csv', 'xls', 'json' and 'pkl' formats.

    Parameters
    ----------
    filename : string, or file-like
        Valid file path or file handle. The string could be a URL.

    Returns
    -------
    The content of the file (DataFrame, array, etc...)

    '''

    if '.csv' in filename:
        content = pd.read_csv(filename)
        content = set_timestamps(content)
        # if we happen to pull the index from a previous save, remove the extra
        # column
        try:
            if all(content['Unnamed: 0'].values == content.index.values):
                content.drop(labels=['Unnamed: 0'], axis=1, inplace=True)
        except:
            pass
    elif '.xls' in filename:
        content = pd.read_excel(filename)
        content = set_timestamps(content)
    elif '.json' in filename:
        content = pd.read_json(filename)
    elif '.pkl' in filename:
        content = pd.read_pickle(filename)
    else:
        content = None
    return content


def to_pickle(obj, filename, cleanMemory=False):
    '''http://stackoverflow.com/questions/7900944/read-write-classes-to-files-in-an-efficent-way'''

    path, filename = path2filename(filename)

    create_dir(path)

    with open(path + filename, "wb") as output:
        pickle.dump(obj, output, pickle.HIGHEST_PROTOCOL)

    if cleanMemory:
        obj = None

    # setting the global object to None requires a return assignment
    return obj


def from_pickle(filename, cleanDisk=False):
    # to deserialize the object
    with open(filename, "rb") as input:
        obj = pickle.load(input)  # protocol version is auto detected

    if cleanDisk:
        os.remove(filename)
    return obj


def path2filename(pathfile):
    '''path2filename takes a path filename string and returns the split between the path and the filename

    if filename is not given, filename = ''
    if path is not given, path = './'
    '''
    path = pathfile[:pathfile.rfind('/') + 1]
    if path == '':
        path = './'
    filename = pathfile[pathfile.rfind('/') + 1:len(pathfile)]
    if '.' not in filename:
        path = pathfile
        filename = ''

    if (filename == '') and (path[len(path) - 1] != '/'):
        path += '/'
    return path, filename


def add_path_string(rootPath='./results', pathString=None):
    rootPath = path2filename(rootPath)[0]
    regEx = '[.<>"!,:;*/ -]'
    if pathString is not None:
        return path2filename(rootPath + re.sub(regEx, '_', pathString))[0]
    else:
        return rootPath


def create_dir(path, dirDict={}):
    '''
    Tries to create a new directory in the given path.
    **create_dir** can also create subfolders according to the dictionnary given as second argument.

    Parameters
    ----------
    path : string
        string giving the path of the location to create the directory, either absolute or relative.
    dirDict : dictionnary, optional (the default is {}, which means that no subfolders will be created)
        Dictionnary ordering the creation of subfolders. Keys must be strings, and values either None or path dictionnaries.

    Examples
    --------

    >>> path = './project'
    >>> dirDict = {'dir1':None, 'dir2':{'subdir21':None}}
    >>> utils.create_dir(path,dirDict)

    will create:

    * *project/dir1*
    * *project/dir2/subdir21*

    in your parent directory.

    '''

    folders = path.split('/')
    folders = [i for i in folders if i != '']
    rootPath = ''
    if folders[0] == 'C:':
        folders = folders[1:]
    count = 0
    for directory in folders:
        count += 1

        # required to handle the dot operators
        if (directory[0] == '.') & (count == 1):
            rootPath = directory
        else:
            rootPath = rootPath + '/' + directory
        try:
            os.makedirs(rootPath)
        # If the file already exists (EEXIST), raise exception and do nothing
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise

    for key in dirDict.keys():
        rootPath = path + "/" + key
        try:
            os.makedirs(rootPath)
        # If the file already exists (EEXIST), raise exception and do nothing
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise
        if dirDict[key] is not None:
            create_dir(rootPath, dirDict[key])


def Walk(root='.', recurse=True, pattern='*'):
    """
        Generator for walking a directory tree.
        Starts at specified root folder, returning files
        that match our pattern. Optionally will also
        recurse through sub-folders.

        Parameters
        ----------
        root : string (default is *'.'*)
            Path for the root folder to look in.
        recurse : bool (default is *True*)
            If *True*, will also look in the subfolders.
        pattern : string (default is :emphasis:`'*'`, which means all the files are concerned)
            The pattern to look for in the files' name.

        Returns
        -------
        generator
            **Walk** yields a generator from the matching files paths.

    """
    for path, subdirs, files in os.walk(root):
        for name in files:
            if fnmatch.fnmatch(name, pattern):
                yield os.path.join(path, name)
        if not recurse:
            break


def scanPath(root='.', recurse=False, pattern='*'):
    '''
    Runs a loop over the :doc:`Walk<relpy.utils.Walk>` Generator
    to find all file paths in the root directory with the given
    pattern. If recurse is *True*: matching paths are identified
    for all sub directories.

    Parameters
    ----------
    root : string (default is *'.'*)
        Path for the root folder to look in.
    recurse : bool (default is *True*)
        If *True*, will also look in the subfolders.
    pattern : string (default is :emphasis:`'*'`, which means all the files are concerned)
        The pattern to look for in the files' name.

    Returns
    -------
    pathList : list
        The list of all the matching files paths.

    '''

    pathList = []
    for path in Walk(root=root, recurse=recurse, pattern=pattern):
        pathList.append(path)
    return pathList


class Timer:

    '''

    Timer that calculates time remaining for a process and the percent complete

    .. todo::

        Ask for details about the usage

    Parameters
    ----------
    nLoops : integer
    numPrints : integer (default is *100*)
    verbose : bool (default is *True*)

    Attributes
    ----------
    nLoops : integer
    numPrints : integer
    verbose : bool
        if *True*, print values when **loop** is called
    count : integer
    elapsed : float
        elapsed time
    est_end : float
        estimated end
    ti : float
        initial time
    tf : float
        current time
    display_amt : integer

    '''

    def __init__(self, nLoops, numPrints=100, verbose=True):

        self.nLoops = nLoops
        self.numPrints = numPrints
        self.verbose = verbose
        self.count = 0
        self.elapsed = 1
        self.est_end = 1
        self.ti = time.time()
        self.display_amt = 1

    def loop(self):
        '''
        Tracks the time in a loop. The estimated time to completion
        can be calculated and if verbose is set to *True*, the object will print
        estimated time to completion, and percent complete.
        Actived in every loop to keep track'''

        self.count += 1
        self.tf = time.time()
        self.elapsed = self.tf - self.ti

        if self.verbose:
            displayAll(self.elapsed, self.display_amt, self.est_end,
                       self.nLoops, self.count, self.numPrints)

    def fin(self):
        print "Elapsed time: %s  :-)" % str(time.time() - self.ti)


def displayAll(elapsed, display_amt, est_end, nLoops, count, numPrints):
    '''Displays time if verbose is true and count is within the display amount'''

    if numPrints > nLoops:
        display_amt = 1
    else:
        display_amt = round(nLoops / numPrints)

    if count % display_amt == 0:

        avg = elapsed / count
        est_end = round(avg * nLoops)

        (disp_elapsed,
         disp_avg,
         disp_est) = timeUnit(int(round(elapsed)),
                              int(round(avg)),
                              int(round(est_end)))

        print "%s%%" % str(round(count / float(nLoops) * 100)), "@" + str(count),

        totalTime = disp_est[0]
        unit = disp_est[1]

        if str(unit) == "secs":
            remain = totalTime - round(elapsed)
            remainUnit = "secs"
        elif str(unit) == "mins":
            remain = totalTime - round(elapsed) / 60
            remainUnit = "mins"
        elif str(unit) == "hr":
            remain = totalTime - round(elapsed) / 3600
            remainUnit = "hr"

        print "ETA: %s %s" % (str(remain), remainUnit)
        print

    return


def timeUnit(elapsed, avg, est_end):
    '''calculates unit of time to display'''
    minute = 60
    hr = 3600
    day = 86400

    if elapsed <= 3 * minute:
        unit_elapsed = (elapsed, "secs")
    if elapsed > 3 * minute:
        unit_elapsed = ((elapsed / 60), "mins")
    if elapsed > 3 * hr:
        unit_elapsed = ((elapsed / 3600), "hr")

    if avg <= 3 * minute:
        unit_avg = (avg, "secs")
    if avg > 3 * minute:
        unit_avg = ((avg / 60), "mins")
    if avg > 3 * hr:
        unit_avg = ((avg / 3600), "hr")

    if est_end <= 3 * minute:
        unit_estEnd = (est_end, "secs")
    if est_end > 3 * minute:
        unit_estEnd = ((est_end / 60), "mins")
    if est_end > 3 * hr:
        unit_estEnd = ((est_end / 3600), "hr")

    return [unit_elapsed, unit_avg, unit_estEnd]


def float_to_string(float_array):
    '''
    Converts an array of  floats into a "readable"
    array of strings (i.e. strings without too many
    caracters, easier to print out on a plot or else).

    Parameters
    ----------
    float_array : array-like
        Array of floats

    Returns
    -------
    string_array : list of strings
        Array of strings made from the floats

    Examples
    --------
    >>> float_ar = [12154.4861652, 2135.355,.0000168498,0.001265,125.35]
    >>> utils.float_to_string(float_ar)
    ['1.22e4', 2135, '1.68e-5', '1.26e-3', '125.3']

    '''

    scalar = np.isscalar(float_array)
    if scalar:
        float_array = [float_array]

    string_array = []
    for f in float_array:
        es = format(f, 'e')
        core, exp = es.split('e')
        core = float(core)
        exp = int(exp)
        if exp > 3 or exp < -1:
            core = round(core, 2)
            sf = str(core) + 'e' + str(exp)
        elif exp > -2 and exp < 3:
            prec = 3 - abs(exp)
            core = round(f, prec)
            sf = str(core)
        else:
            sf = int(f)
        string_array.append(sf)

    if scalar:
        return string_array[0]
    else:
        return string_array



'''parallel contains best practices for parallel processing turntable.utilsities
	Authors: Jason Shiverick <jshiverick@teslamotors.com>
			 Carlo Torniai <ctorniai@teslamotors.com>
'''

# import multiprocessing tools (source:
# http://chriskiehl.com/article/parallelism-in-one-line/)
import multiprocessing as mp  # for processes
from multiprocessing.dummy import Pool as ThreadPool  # for threads

import time
import pickle
import os
import shutil

import turntable.utils


def series(sequence, function, *args, **kwargs):
    '''
    series(sequence[, sequence, ...], function) -> list

    Parameters
    ----------
    sequence : list of items to process
    function : function to process each item


    Return a list of the results of applying the function to the items of
    the argument sequence(s).  If more than one sequence is given, the
    function is called with an argument list consisting of the corresponding
    item of each sequence, substituting None for missing values when not all
    sequences have the same length.  If the function is None, return a list of
    the items of the sequence (or a list of tuples if more than one sequence).
    '''

    if 'verbose' in kwargs.keys():
        verbose = kwargs['verbose']
    else:
        verbose = True

    results = []
    timer = turntable.utils.Timer(nLoops=len(sequence), numPrints=15, verbose=verbose)
    for subject in sequence:
        results.append(function(subject, *args, **kwargs))
        timer.loop()
    timer.fin()
    return results


def batch(sequence, function, processes=None, batch_size=None, quiet=False,
          kwargs_to_dump=None, args=None, **kwargs):
    '''runs a process in series for batches where each batch is run parallel over the number of processes specified.
    Allows for aggregation and incremental storage and runs the in memory processes

    sequence is an array of car instances
    batch_size is the specified size of each batch. defaults to len(sequence)/processes
    processes is the number of processes to use
            Defaults to the number processes multiprocessing calculates up to a max of 20'''

    if processes is None:
        # default to the number of processes, not exceeding 20 or the number of
        # subjects
        processes = min(mp.cpu_count(), 20, len(sequence))

    if batch_size is None:
        # floor divide rounds down to nearest int
        batch_size = max(len(sequence) // processes, 1)
    print 'size of each batch =', batch_size

    mod = len(sequence) % processes
    # batch_list is a list of cars broken in to batch size chunks
    batch_list = [sequence[x:x + batch_size]
                  for x in xrange(0, len(sequence) - mod, batch_size)]
    # remainder handling
    if mod != 0:
        batch_list[len(batch_list) - 1] += sequence[-mod:]
    print 'number of batches =', len(batch_list)

    # #if the sequence is shorter than batch_size
    # #the system will run parallel to completion in one iteration
    # if batch_size >= len(sequence):
    # 	verbose = False
    # else:
    # 	verbose = (not quiet)

    # timer = turntable.utils.Timer(nLoops = len(batch_list), numPrints = len(batch_list), verbose = verbose)
    # #run the list of batches in series
    # tic = time.time()
    # returnList = []
    # for batch in batch_list:
    # 	#Process each batch in parallel
    # 	batch = thread(function = function, sequence = batch, processes = processes, runSeries = False, quiet = quiet)
    # 	[returnList.append(subject) for subject in batch]#bring all previously runs into a single list
    # 	timer.loop()

    # New args
    if args is None:
        args = function
    else:
        if isinstance(args, tuple) == False:
            args = (args,)
        args = (function,) + args

    # Applying the mp function w/ or w/o dumping using the custom operator
    # function
    if kwargs_to_dump is None:
        res = parallel(
            batch_list,
            new_function_batch,
            processes=processes,
            args=args,
            **kwargs)
    else:
        res = process_dump(
            batch_list,
            new_function_batch,
            kwargs_to_dump,
            processes=processes,
            args=args,
            **kwargs)

    returnList = []
    for l in res:
        returnList += l

    # toc = time.time()
    # elapsed = toc-tic
    # if quiet is False:
    # 	if processes is None:
    # 		print "Total Elapsed time: %s  :-)" % str(elapsed)
    # 	else:
    # print "Total Elapsed time: %s  on %s processes :-)" %
    # (str(elapsed),str(processes))

    return returnList



def new_function_batch(sequence, function, *args, **kwargs):

    proc_name = mp.current_process().name
    print proc_name + ' starts'
    print 'Processing ' + str(len(sequence)) + ' arguments'

    results = []
    timer = turntable.utils.Timer(nLoops=len(sequence), numPrints=10, verbose=False)
    for subject in sequence:
        results.append(function(subject, *args, **kwargs))
        timer.loop()
    timer.fin()
    # return results

    # proc_name = mp.current_process().name
    # print proc_name+' starts'
    # tic = time.time()
    # res = []
    # print 'Processing '+str(len(main_arg_l))+' arguments'
    # for main_arg in main_arg_l:
    # 	res.append(function(main_arg,*args,**kwargs))

    # print proc_name+' ends in '+str(time.time()-tic)
    return results


def thread(
        function, sequence, cores=None, runSeries=False, quiet=False):
    '''sets up the threadpool with map for parallel processing'''

    # Make the Pool of workes
    if cores is None:
        pool = ThreadPool()
    else:
        pool = ThreadPool(cores)

    # Operate on the list of subjects with the requested function
    # in the split threads
    tic = time.time()
    if runSeries is False:
        try:
            results = pool.map(function, sequence)
            # close the pool and wiat for teh work to finish
            pool.close()
            pool.join()
        except:
            print 'thread Failed... running in series :-('
            results = series(sequence, function)
    else:
        results = series(sequence, function)
    toc = time.time()
    elapsed = toc - tic

    if quiet is False:
        if cores is None:
            print "Elapsed time: %s  :-)" % str(elapsed)
        else:
            print "Elapsed time: %s  on %s threads :-)" % (str(elapsed), str(cores))
    # Noes:
    # import functools
    # abc = map(functools.partial(sb.dist, distName = 'weibull'), wbldfList)

    return results


def parallel(sequence, function, processes=None, args=None, **kwargs):

    if processes is None:
        # default to the number of cores, not exceeding 20
        processes = min(mp.cpu_count(), 20)
    print "Running parallel process on " + str(processes) + " cores. :-)"

    pool = mp.Pool(processes=processes)
    PROC = []
    tic = time.time()
    for main_arg in sequence:
        if args is None:
            ARGS = (main_arg,)
        else:
            if isinstance(args, tuple) == False:
                args = (args,)
            ARGS = (main_arg,) + args
        PROC.append(pool.apply_async(function, args=ARGS, kwds=kwargs))
    #RES = [p.get() for p in PROC]
    RES = []
    for p in PROC:
        try:
            RES.append(p.get())
        except Exception as e:
            print "shit happens..."
            print e
            RES.append(None)
    pool.close()
    pool.join()

    toc = time.time()
    elapsed = toc - tic
    print "Elapsed time: %s  on %s processes :-)" % (str(elapsed), str(processes))

    return RES


def new_function_dumping(
        args_to_load_names, function, main_arg, *args, **kwargs):

    for name in args_to_load_names:
        f_toread = open('./temp_pickle/' + name + '.pkl', 'r')
        kwargs[name] = pickle.load(f_toread)
        f_toread.close()
    if args is not None:
        res = function(main_arg, *args, **kwargs)
    else:
        res = function(main_arg, **kwargs)
    return res


def process_dump(
        sequence,
        function,
        kwargs_to_dump,
        processes=None,
        args=None,
        **kwargs):

    if processes is None:
        # default to the number of cores, not exceeding 20
        processes = min(mp.cpu_count(), 20)
    print "Running parallel process on " + str(processes) + " cores. :-)"

    tic = time.time()
    turntable.utils.create_dir('./temp_pickle/')
    for name, obj in kwargs_to_dump.iteritems():
        f_pkl = open('./temp_pickle/' + name + '.pkl', 'w+')
        pickle.dump(obj, f_pkl)
        f_pkl.close()
    args_to_load_names = kwargs_to_dump.keys()

    def arg_list(main_arg, args):
        if args is None:
            ARGS = (args_to_load_names, function, main_arg)
        else:
            if isinstance(args, tuple) == False:
                args = (args,)
            ARGS = (args_to_load_names, function, main_arg,) + args
        return ARGS

    pool = mp.Pool(processes=processes)
    PROC = []

    for main_arg in sequence:
        ARGS = arg_list(main_arg, args)
        PROC.append(
            pool.apply_async(new_function_dumping, args=ARGS, kwds=kwargs))

    #RES = [p.get() for p in PROC]
    RES = []
    for p in PROC:
        try:
            RES.append(p.get())
        except Exception as e:
            print "shit happens..."
            print e
            RES.append(None)
    print 'Closing processes...'
    pool.close()
    pool.join()

    if os.path.exists('./temp_pickle/'):
        shturntable.utils.rmtree('./temp_pickle/')

    toc = time.time()
    elapsed = toc - tic
    print "Elapsed time: %s on %s processes :-)" % (str(elapsed), str(processes))

    return RES



print '''
Installing turntable....


                         _
_._ _..._ .-',     _.._(`))
'-. `     '  /-._.-'    ',/
   )         \            '.
  / _    _    |             \
|  a    a    /              |
\   .-.                     ;  
  '-('' ).-'       ,'       ;
     '-;           |      .'
        \           \    /
        | 7  .__  _.-\   \
        | |  |  ``/  /`  /
       /,_|  |   /,_/   /
          /,_/      '`-'
'''
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

from version import *
import sys


config = {
    'description': 'TURNTABLE provides a functional framework for parallel processing DataFrames',
    'author': 'Jason Shiverick',
    'url': 'https://github.com/jshiv/turntable',
    'download_url': 'https://github.com/jshiv/turntable',
    'author_email': 'jshiv00@gmail.com',
    'license':'MIT',
    'version': get_git_version(),
    
    'install_requires': [
    'pandas',
    ],

    'packages': find_packages(),#['turntable'],
    'scripts': [],
    'name': 'turntable'
}



print "system is: "+sys.platform
print ''
print "installing turntable dependencies... "
print config['install_requires']

setup(**config)


"""This is the docstring for the example.py module.  Modules names should
have short, all-lowercase names.  The module name may have underscores if
this improves readability.

Every module should have a docstring at the very top of the file.  The
module's docstring may extend over multiple lines.  If your docstring does
extend over multiple lines, the closing three quotation marks must be on
a line by itself, preferably preceeded by a blank line.

"""
# uncomment the following line to use python 3 features
#from __future__ import division, absolute_import, print_function

import shutil
import sys
import os  # standard library imports first

# Do NOT import using *, e.g. from numpy import *
#
# Import the module using
#
#   import numpy
#
# instead or import individual functions as needed, e.g
#
#  from numpy import array, zeros
#
# If you prefer the use of abbreviated module names, we suggest the
# convention used by NumPy itself::


import pandas as pd

import turntable.utils
#import parallel
import traceback


# These abbreviated names are not to be used in docstrings; users must
# be able to paste and execute docstrings after importing only the
# numpy module itself, unabbreviated.


class RecordPress(object):
    '''This class auto-seralizes any attributes assigned to an instance and clears them from memmory
    when an attribute is called via the dot operator, it is read from disk

    Parameters
    ----------
    pickle : Boolean [True]
        if False, the instance will behave as a normal class
    pickle_path : string ['./tmp']
        the path underwhich the files will be stored

    Methods
    -------
    clean_disk()
        deletes all files stored by the instance
    clean_memmory()
        sets the in memory attribute values to None reducing the memory footprint


    Examples
    --------
    This class can be encapsulated to be used elsewhere

    >>> class NewClass(RecordPress):
    >>>
    >>>  def __init__(self, pickle = True, pickle_path = './tmp'):
    >>>     self.pickle = pickle
    >>>     self.class_path = turntable.utils.path2filename(pickle_path+'/'+self.__class__.__name__)[0]
    >>>     self.pickles = []
    >>>
    >>> newClass = NewClass()
    >>> newClass.x = 10
    >>> y = newClass.x
    >>> newClass.clean_disk()
    '''
    def __init__(self, pickle=True, pickle_path='./tmp'):
        '''
        Parameters
        ----------
        pickle : pickle the class properties
        pickle_path : where to pickle
        '''

        self.pickle = pickle  # required for over setattr override function
        self.class_path = turntable.utils.path2filename(
            pickle_path + '/' + self.__class__.__name__)[0]
        self.pickles = []

    def __getattribute__(*args):
        # ignore these standard calls to class properties
        ignore_args_list = ['__dict__', '__members__', '__methods__',
                            '__class__', 'trait_names', '_getAttributeNames']
        if any([attr in args[1]
                for attr in ['pickle', 'class_path', 'pickles'] + ignore_args_list]):
            # if the name of the args matches the standard or required names,
            # recurse normally
            return object.__getattribute__(*args)
        else:
            self = args[0]  # 0th index is the class instance
            name = args[1]  # 1st index is the attribute lable requested
            # if this is a attribute from the pickles list, grab it from disk
            if self.pickle & (name in self.pickles):
                # print "Reading disk: ", self.class_path + name + '.pkl'
                return turntable.utils.from_pickle(
                    filename=self.class_path + name + '.pkl', cleanDisk=False)
            else:  # otherwise return the attribute from the class instance
                # print "Reading memory: ", name
                return object.__getattribute__(*args)

    def __setattr__(self, name, value):
        # first set the attribute to the instance of the class
        object.__setattr__(self, name, value)

        # test for the pickle command
        if self.pickle:
            # if the name of the assignend attribute is not a required name, go
            # ahead and pickle it
            if any(
                    [attr in name for attr in ['pickle', 'class_path', 'pickles']]) == False:
                return_value = turntable.utils.to_pickle(
                    obj=value,
                    filename=self.class_path +
                    name +
                    '.pkl',
                    cleanMemory=True)
                # return_value = None #for testing memmory
                self.__dict__[name] = return_value
                # turntable.utils.to_pickle(obj = value, filename = self.class_path + name + '.pkl', cleanMemory = False) #use this line to not default memory clearing
                # add the name to the list of pickled attributes, so we know
                # which attributes have been pickled
                self.pickles.append(name)
                self.pickles = list(set(self.pickles))
                # print "writing to disk: ", self.class_path + name + '.pkl'
            else:
                pass
                # print "writing to memory: ", name
        else:
            pass
            print  # "writing to memory: ", name

    def clean_memory(self):
        '''sets all attribute values from the attribute list pickles to None
        -this is the default behaviure so this method is redundent'''
        for key in self.pickles:
            self.__dict__[key] = None

    def clean_disk(self):
        '''clean removes all files and folders under the class_path directory'''
        shutil.rmtree(self.class_path)


class RecordSetter:

    '''RecordSetter provides a simple interface for initalizing arguments passed in kwargs
    and a runMethod method for running a class method by name


    Parameters
    ----------
    kwargs : name : value

    Examples
    --------
    RecordSetter is a general python class that assigns **kwargs as instances of its self.

    >>> obj = RecordSetter( name = 'me')
    >>> print obj.name
    me
    '''

    def __init__(self, **kwargs):
        self.set_attributes(kwargs)

    def set_attributes(self, kwargs):
        '''
        Initalizes the given argument structure as properties of the class
        to be used by name in specific method execution.

        Parameters
        ----------
        kwargs : dictionary
            Dictionary of extra attributes,
            where keys are attributes names and values attributes values.

        '''
        for key, value in kwargs.items():
            setattr(self, key, value)
            # print key, value

    def run_method(self, method):
        '''call a specied method by name using runMethod'''
        methodToCall = getattr(self, method)
        result = methodToCall()


class SeriesLoader(object):
    '''This class assignes given properties to a special pandas.Series propertie

    self.series

    Parameters
    ----------
    series : all atributes of the class get added to an internal pandas series


    Examples
    --------
    This class can be encapsulated to be used elsewhere

    >>> series_loader = SeriesLoader()
    >>> series_loader.one = 'one'
    >>> print series_loader.series
    '''
    def __init__(self):
        pass

    def __setattr__(self, name, value):
        # first set the attribute to the instance of the class
        object.__setattr__(self, name, value)
        # if the name of the assignend attribute is not a required name, go
        # ahead and pickle it
        blacklisted_keys = ['series']
        if any([attr in name for attr in blacklisted_keys]) == False:
            if 'series' in self.__dict__.keys():
                self.series[name] = value
            else:
                self.series = pd.Series(dict(
                    [(key, value) for key, value in self.__dict__.items() if key not in blacklisted_keys]))


class Record(RecordSetter, SeriesLoader):
    '''
    Record is a container object with a the special propertie "series"

    any propertie added to Record will also be added to the pandas.Series

    Properties
    ----------
    series : pandas.Series
        container for parameters set to the instance

    Methods
    -------
    set_attributes : assigns items of a dictionary to the class and to the series parameter
    runMethod : runs a method by a string call

    Examples
    --------
    lets see how we can add a propertie to the record object

    >>> record = Record(first_item = 'one')
    >>> record.second_item = 'two'
    >>> print record.series
    '''
    def __init__(self, **kwargs):

        self.set_attributes(kwargs)


def load_record(index_series_touple, kwargs):
    '''
    generates an instance of Record() from a touple of the form (index, pandas.Series) with associated parameters kwargs

    Paremeters
    ----------
    index_series_touple : (index, pandas.Series)
    kwargs : dict
        aditional arguments

    Returns
    -------
    Record()
    '''
    index_record = index_series_touple[0]
    series = index_series_touple[1]
    record = Record()
    record.series = series
    record.index_record = index_record
    record.set_attributes(kwargs)
    return record


def build_collection(df, **kwargs):
    '''
    build_collection generates a list of Record objects given a population DataFrame.
    Each record instance has a series attribute which is a pandas.Series of the same attributes in df.
    Optional datasets can be passed in through kwargs which
    will be included by the name of each object

    parameters
    ----------
    df : pandas.DataFrame
        each record represents a one individual from a population
    kwargs : alternate arguments to be saved by name to the series of each object

    Returns
    -------
    fleet : list
        list of Record() objects
    '''
    print 'Generating the Record Collection...'
    print ' '

    df['index_original'] = df.index
    df.reset_index(drop=True, inplace=True)

    if pd.__version__ >= '0.15.0':
        d = df.T.to_dict(orient='series')
    else:
        d = df.T.to_dict(outtype='series')

    collection = [load_record(item, kwargs) for item in d.items()]

    #collection = parallelBox.batch(d.items(), load_record, nb_proc=None, batchSize=None, quiet=True, kwargs_to_dump=None, args = kwargs)

    return collection


def collection_to_df(collection):
    return pd.concat([record.series for record in collection], axis=1).T


def catch(fcn, *args, **kwargs):
    '''try:
          retrun fcn(*args, **kwargs)
       except:
          print traceback
            if 'spit' in kwargs.keys():
                return kwargs['spit']


    Parameters
    ----------
    fcn : function
    *args : unnamed parameters of fcn
    **kwargs : named parameters of fcn
        spit : returns the parameter named return in the exception

    Returns
    -------
    The expected output of fcn or prints the exception traceback'''
    try:
        # remove the special kwargs key "spit" and use it to return if it
        # exists
        spit = kwargs.pop('spit')
    except:
        spit = None

    try:
        results = fcn(*args, **kwargs)
        if results is not None:
            return results
    except:
        # traceback.print_exc()
        print traceback.format_exc()
        if spit is not None:
            return spit


def foo(var1, var2, long_var_name='hi'):
    r"""A one-line summary that does not use variable names or the
    function name.

    Several sentences providing an extended description. Refer to
    variables using back-ticks, e.g. `var`.

    Parameters
    ----------------
    var1 : array_like
        Array_like means all those objects -- lists, nested lists, etc. --
        that can be converted to an array.  We can also refer to
        variables like `var1`.
    var2 : int
        The type above can either refer to an actual Python type
    Long_variable_name : {'hi', 'ho'}, optional
        Choices in brackets, default first when optional.

    Returns
    ----------------
    type
        Explanation of anonymous return value of type ``type``.
    describe : type
        Explanation of return value named `describe`.
    out : type
        Explanation of `out`.

    Other Para
    ----------------
    only_seldom_used_keywords : type
        Explanation
    common_parameters_listed_above : type
        Explanation

    Raises
    ------
    BadException
        Because you shouldn't have done that.

    See Also
    --------
    otherfunc : relationship (optional)
    newfunc : Relationship (optional), which could be fairly long, in which
              case the line wraps here.
    thirdfunc, fourthfunc, fifthfunc

    Notes
    -----
    Notes about the implementation algorithm (if needed).

    This can have multiple paragraphs.

    You may include some math:

    .. math:: X(e^{j\omega } ) = x(n)e^{ - j\omega n}

    And even use a greek symbol like :math:`omega` inline.

    References
    ----------
    Cite the relevant literature, e.g. [1]_.  You may also cite these
    references in the notes section above.

    .. [1] O. McNoleg, "The integration of GIS, remote sensing,
       expert systems and adaptive co-kriging for environmental habitat
       modelling of the Highland Haggis using object-oriented, fuzzy-logic
       and neural-network techniques," Computers & Geosciences, vol. 22,
       pp. 585-588, 1996.

    Examples
    --------
    These are written in doctest format, and should illustrate how to
    use the function.

    >>> a=[1,2,3]
    >>> print [x + 3 for x in a]
    [4, 5, 6]
    >>> print "a\n\nb"
    a
    b

    """

    pass
